/***********************************************************
 *File: Engine.cpp
 *Author: Peter Asplund
 *Created: 2004-11-08
 *Description: Simulates the engine. 
 *             Receives boolean wether to accelerate or 
 *             not. If true, it calculates the speed. 
 *             It also handles gears and brakes.
 **********************************************************/

#include "include/engine.h"
#include <math.h>
#include <vector>
#include <iostream>

using namespace std;

namespace engine {
  
  /******************************************
  *Function: Constructor
  *Description: Initiate variables
  ******************************************/
  Engine::Engine() : isAccelerating(false), angleVel(0), speed(0), torque(0), brakeFriction(0), currentGear(2), maxSpeed(1.0) {

    dX1 = 0;

    /*********************************
     *Initiate transmission
     *Each position in vector
     *represents one gear.
     *4 Forward and
     *2 Backwards
     ********************************/
    
    transmission.push_back(-1.0/0.1923);
    transmission.push_back(-1.0/0.0962);
    transmission.push_back(0.0);
    transmission.push_back(1.0/0.0962);
    transmission.push_back(1.0/0.1923);
    transmission.push_back(1.0/0.2885);
    transmission.push_back(1.0/0.4487);

  }

  /******************************************
   *Function: setGearUp
   *Arguments: NULL
   *Returns: True if shifted gear, else false
   *Description: Increases currentGear if the new gear is possible
   *             It is impossible to shift to drive
   *             if tank is going in reverse to fast.
   *****************************************/
  bool Engine::setGearUp() {

    if (currentGear < 6) {
    
      if (currentGear < 1) {

	currentGear++;
	return true;
	
      }
      else if (speed > -0.5) {
	
	currentGear ++;
	return true;

      }
    }

    else {

      return false;

    }
    
  }

  /******************************************
   *Function: setGearDown
   *Arguments: NULL
   *Returns: True if shifted gear, else false
   *Description: Decreases currentGear if the new gear is possible
   *             It is impossible to shift to reverse
   *             if tank is going forward to fast.
   *****************************************/
  bool Engine::setGearDown() {

    if (currentGear > 0) {
      
      if (currentGear > 3) {
	
	currentGear--;
	return true;
      
      }
      else if (speed < 0.5) {
	
	currentGear--;
	return true;

      }
   
    }

    else {

    return false;

    }
    
  }

  /******************************************
   *Function:    setMaxSpeed
   *Arguments:   the percentage of the engine to use
   *Returns:     NULL
   *Description: Sets the maximum percentage to use of the engines power (0.0 - 1.0)
   *****************************************/
  void Engine::setMaxSpeed(float newSpeed) {
    
    if (newSpeed <= 1.0 && newSpeed >= 0.0) {
      if (newSpeed = 0)
	maxSpeed = 1.0;
      else
	maxSpeed = newSpeed;
    }

    else {} //Do nothing

  }

  /******************************************
   *Function:   calculateAcceleration
   *Arguments:  bool accelerate
   *            bool breaking
   *            float stepSize (1/fps)
   *Returns:    NULL
   *Description:Calculates acceleration/speed 
   *            by using a differential eq. 
   *            (state space description)
   *            and Improved Euler approximation
   *****************************************/
  void Engine::calculateAcceleration (bool& accelerate, bool& breaking, float& stepSize) {
    
    isAccelerating = accelerate;
    isBreaking = breaking;

    if (!isAccelerating || currentGear == 2)
      torque = 0;

    if (isAccelerating && currentGear != 2) {

      if (currentGear == 3 && angleVel <= 44.0)
	torque = 11520;

      if (currentGear == 4 && angleVel <= 44.0)
	torque = 11520;
      
      else
	torque = 4800;// * maxSpeed;

      if (currentGear != 3)
	torque = torque * ( 0.5 + (0.5 * angleVel) / 43.0); //straight line ekvation for linear torque depending on angleVel.
	
    }

    if (isBreaking)
	brakeFriction = 20000;        //Friction on brake if breaking

    if (!isBreaking)
      brakeFriction = 0;
  
    /*****************************************************
     *Calculate acceleration (from state space description)
     ****************************************************/
    dX1 = (1.0 / CHASSIMASS) * ( torque * transmission.at(currentGear) - speed * ( INTERNALFRICTION * transmission.at(currentGear) * transmission.at(currentGear) + GROUNDFRICTION + brakeFriction));
    
    /***********************************
     *Improved Euler (x(h+1))
     ***********************************/
    uSpeed += stepSize * dX1;
    
    /***********************************
     *calculate dX(h+1) (dU)
     **********************************/
    dU1 = (1.0 / CHASSIMASS) * ( torque * transmission.at(currentGear) - uSpeed * ( INTERNALFRICTION * transmission.at(currentGear) * transmission.at(currentGear) + GROUNDFRICTION + brakeFriction));
        
    /***********************************
     *Improved Euler approximation
     ***********************************/
    speed += stepSize * dX1;//; + dU1)/2;
    drivingForce = CHASSIMASS * dX1;// + dU1)/2;
    angleVel = speed * transmission.at(currentGear);
    //cout << angleVel*30/M_PI << ":" << speed << endl;
  }

  /******************************************
   *Function: getSpeed
   *Arguments: NULL
   *Returns: the speed of the tank
   *****************************************/
  double Engine::getSpeed() const {

    return speed;

  }

  /******************************************
   *Function:    getForce
   *Arguments:   NULL
   *Returns:     the driving force of the tank
   *****************************************/ 
  double Engine::getForce() const {

    return drivingForce;

  }

  /*****************************************
   *Function:    getRPM
   *Arguments:   NULL
   *Returns:     The RPM of the tank
   *Description: Calculates the RPM from the
   *             angular velocity given by the
   *             state space description
   ****************************************/
  double Engine::getRPM() const {

    return (angleVel);//*30/M_PI);

  }

 /******************************************
   *Function:    getGear
   *Arguments:   NULL
   *Returns:     the current gear
   *Description: -2 is to adjust the output to 
   *             the correct gear, and not 
   *             the position in the vector
   *****************************************/  
  int Engine::getGear() const {

    return (currentGear-2);

  }

}
